from datetime import date
from urllib.parse import urlencode, urljoin

import scrapy
from dateutil import rrule
from scrapy.loader import ItemLoader

from ..items import WaterConsumptionItem


class WaterConsumptionSpider(scrapy.Spider):
    name = "water_consumption"

    def __init__(self, bvk_username, bvk_password, date_from, date_to=None, cb_item_scraped=None, *args, **kwargs):
        """ Get the water consumption from BVK. It will scrape the BVK website
            using the provided username/password and get daily consumption values
            for the given interval specified by `date_from` and `date_to`
            (which are datetime.date-compatible objects). If `date_to` is
            not provided it will scrape data starting from `date_from` till
            today.

            The `cb_item_scraped` is a callback accepting one argument "item"
            that will be called for each item (consumption data for the given
            month) scraped.
        """
        super().__init__(*args, **kwargs)

        # BVK username/password
        self._bvk_username = bvk_username
        self._bvk_password = bvk_password

        # Year/month to scrape
        self._date_from = date_from
        self._date_to = date_to if date_to is not None else date.today()

        # Output callback - will be called for each item scraped
        self._cb_item_scraped = cb_item_scraped

    def start_requests(self):
        url = "https://zis.bvk.cz/"
        yield scrapy.Request(url=url, callback=self.handle_login)

    def handle_login(self, response):
        return scrapy.FormRequest.from_response(
            response,
            formid="aspnetForm",
            headers={"X-MicrosoftAjax": "Delta=true"},
            formdata={
                "ctl00$lvLoginForm$LoginDialog1$edEmail": self._bvk_username,
                "ctl00$lvLoginForm$LoginDialog1$edPassword": self._bvk_password,
                "ctl00$lvLoginForm$LoginDialog1$btnLogin": "Login",
            },
            callback=self.handle_login_response,
        )

    def handle_login_response(self, response):
        if "pageRedirect" in response.text:
            main_info_url = urljoin(response.url, "/Userdata/MainInfo.aspx")
            return scrapy.Request(url=main_info_url, callback=self.handle_main_info_response)
        else:
            raise Exception(f"Login to BVK failed (status {response.status}): {response.text}")

    def handle_main_info_response(self, response):
        suezsmartsolutions_login_url = response.css(
            "a#ctl00_ctl00_ContentPlaceHolder1_UserDataContentPlaceHolder_btnPortalEmis"
        ).attrib["href"]
        return scrapy.Request(url=suezsmartsolutions_login_url, callback=self.handle_suezsmartsolutions_login_response)

    def handle_suezsmartsolutions_login_response(self, response):
        # Iterate over years/months between `date_from` and `date_to` and get
        # the daily consumption for the respective months
        #
        # We need to set the day in `dtstart` to the first day of month because
        # otherwise if the day is greater than the day in `until` the last
        # month would not be included. For example for dtstart 2020-03-09 and
        # until 2020-08-01 would skip 2020-08 because 2020-08-09 comes after
        # 2020-08-01.
        for dt in rrule.rrule(rrule.MONTHLY, dtstart=self._date_from.replace(day=1), until=self._date_to):
            qs = urlencode(
                {
                    "Affichage": "ConsoJour",  # consumption by day
                    "Annee": dt.year,  # for this year
                    "Mois": dt.month,  # and this month
                }
            )
            consumption_url = urljoin(response.url, f"/eMIS.SE_BVK/Site_Energie.aspx?{qs}")
            yield scrapy.Request(
                url=consumption_url,
                callback=self.handle_suezsmartsolutions_consumption_response,
                cb_kwargs={
                    "year": dt.year,
                    "month": dt.month,
                    "not_before": self._date_from,
                    "not_after": self._date_to,
                },
            )

    def handle_suezsmartsolutions_consumption_response(self, response, year, month, not_before, not_after):
        table_rows = response.css("table#ctl00_PHZonePrincipale_ctl01_TableTableau tr")
        for row in table_rows[1:]:  # Skip table header
            loader = ItemLoader(item=WaterConsumptionItem(), selector=row)
            loader.add_css("date", "td.TableauEnergieLabel::text")
            loader.add_css("consumption", "td.TableauEnergie span::text")
            item = loader.load_item()

            # Check if the date in consumption data is within our time frame;
            # if it is, yield the consumption data, otherwise continue with the
            # next item
            if not_before <= item["date"] <= not_after:
                self._cb_item_scraped(item)
                yield item



# -*- coding: utf-8 -*-

# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html

from datetime import datetime

import scrapy
from scrapy.loader.processors import TakeFirst


def parse_date(values):
    return [datetime.strptime(value, "%m/%d/%Y").date() for value in values]


def parse_consumption(values):
    return [int(value) for value in values]


class WaterConsumptionItem(scrapy.Item):
    date = scrapy.Field(input_processor=parse_date, output_processor=TakeFirst())
    consumption = scrapy.Field(input_processor=parse_consumption, output_processor=TakeFirst())



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    install_requires = [line for line in fh if line and line[0] not in "#-"]

setuptools.setup(
    name="python-bvk",
    version="0.1.1",
    author="Dan Keder",
    author_email="dan.keder@protonmail.com",
    description="Python library for tracking water consumption from BVK (Brnenske vodarny a kanalizace, bvk.cz)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dankeder/python-bvk",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



from twisted.internet import reactor
from scrapy.settings import Settings
from scrapy.crawler import CrawlerRunner

from .spiders.water_consumption import WaterConsumptionSpider


class Bvk:
    def __init__(self, bvk_username, bvk_password, log_enabled=None, log_level=None):
        self._bvk_username = bvk_username
        self._bvk_password = bvk_password

        # NOTE: We create settings "manually", not using
        # "scrapy.utils.project.get_project_settings" so we don't have to
        # bundle scrapy.cfg with the bvk package.
        self._settings = Settings()
        self._settings.setmodule('bvk.settings', priority='project')

        # Override settings if needed
        if log_enabled is not None:
            self._settings["LOG_ENABLED"] = log_enabled
        if log_level is not None:
            self._settings["LOG_LEVEL"] = log_level

    def getWaterConsumption(self, date_from, date_to=None):
        """ Get water consumption for the specified time period.  If `date_to`
            is not specified return consumption data from `date_from` till today.

            `date_from` and `date_to` must be datetime.date-compatible objects.
        """
        items = []

        def _item_scraped(item):
            items.append(item)

        runner = CrawlerRunner(self._settings)
        deferred = runner.crawl(
            WaterConsumptionSpider,
            bvk_username=self._bvk_username,
            bvk_password=self._bvk_password,
            date_from=date_from,
            date_to=date_to,
            cb_item_scraped=_item_scraped,
        )
        deferred.addBoth(lambda _: reactor.stop())
        reactor.run(installSignalHandlers=False)  # the script will block here until the crawling is finished

        return {item["date"].isoformat(): item["consumption"] for item in items}
